@@ -32,6 +32,15 @@ let let_ins _loc bindings expr =
     <:expr< let $binding$ in $expr$ >>
   ) bindings expr
 
+let alias_or_fun expr fct =
+  let is_id =
+    match expr with
+    | <:expr< $id:_$ >> -> true
+    | _ -> false
+  in
+  if is_id then expr else fct
+;;
+
 (* Generators for the binary protocol *)
 
 (* Generates the signature for binary protocol writers *)
@@ -347,7 +356,7 @@ module Generate_bin_size = struct
       in
       match loop _loc rhs with
       | `Fun fun_expr when !is_nil -> fun_expr
-      | `Fun fun_expr -> <:expr< fun v -> $fun_expr$ v >>
+      | `Fun fun_expr -> alias_or_fun fun_expr <:expr< fun v -> $fun_expr$ v >>
       | `Match matchings -> <:expr< fun [ $matchings$ ] >>
     in
     let tparam_cnvs = List.map ((^) "_size_of_" *** Gen.get_tparam_id) tps in
@@ -478,8 +487,8 @@ module Generate_bin_write = struct
       | <:ctyp< $tp1$ | $tp2$ >> -> <:match_case< $loop tp1$ | $loop tp2$ >>
       | <:ctyp< `$cnstr$ >> ->
           <:match_case<
-            `$cnstr$ as v ->
-              Bin_prot.Write.bin_write_variant_tag buf ~pos v
+            `$cnstr$ ->
+              Bin_prot.Write.bin_write_variant_int buf ~pos $`int:Pa_type_conv.hash_variant cnstr$
           >>
       | <:ctyp< `$cnstr$ of $tp$ >> ->
           let write_args =
@@ -488,9 +497,9 @@ module Generate_bin_write = struct
             | `Match matchings -> <:expr< match args with [ $matchings$ ] >>
           in
           <:match_case<
-            `$cnstr$ args as v ->
+            `$cnstr$ args ->
               let pos =
-                Bin_prot.Write.bin_write_variant_tag buf ~pos v
+                Bin_prot.Write.bin_write_variant_int buf ~pos $`int:Pa_type_conv.hash_variant cnstr$
               in
               $write_args$
           >>
@@ -605,7 +614,8 @@ module Generate_bin_write = struct
       in
       match loop _loc rhs with
       | `Fun expr when !is_nil -> <:expr< fun _buf ~pos:_ _v -> $expr$ >>
-      | `Fun fun_expr -> <:expr< fun buf ~pos v -> $fun_expr$ buf ~pos v >>
+      | `Fun fun_expr ->
+        alias_or_fun fun_expr <:expr< fun buf ~pos v -> $fun_expr$ buf ~pos v >>
       | `Match matchings -> <:expr< fun buf ~pos -> fun [ $matchings$ ] >>
     in
     let tparam_cnvs = List.map ( (^) "_write_" *** Gen.get_tparam_id) tps in
@@ -627,7 +637,7 @@ module Generate_bin_write = struct
         let call =
           Gen.apply _loc <:expr< $lid:size_name$ >> tparam_size_exprs
         in
-        <:expr< fun v -> $call$ v >>
+        alias_or_fun call <:expr< fun v -> $call$ v >>
       in
       let tparam_write_exprs =
         List.map (fun tp ->
@@ -641,7 +651,7 @@ module Generate_bin_write = struct
         let call =
           Gen.apply _loc <:expr< $lid:write_name$ >> tparam_write_exprs
         in
-        <:expr< fun buf ~pos v -> $call$ buf ~pos v >>
+        alias_or_fun call <:expr< fun buf ~pos v -> $call$ buf ~pos v >>
       in
       let write =
         <:expr<
@@ -831,7 +841,7 @@ module Generate_bin_read = struct
     let atoms_only = ref true in
     let code =
       let mk_check_vint mcs =
-        <:expr< match Bin_prot.Common.variant_of_int vint with [ $mcs$ ] >>
+        <:expr< match vint with [ $mcs$ ] >>
       in
       let mk_try_next_expr call next_expr =
         <:expr<
@@ -849,13 +859,13 @@ module Generate_bin_read = struct
             | `None -> raise_nvm
       and loop_one next t = function
         | <:ctyp< `$cnstr$ >> ->
-            let this_mc = <:match_case< `$cnstr$ as tag -> tag >> in
+            let this_mc = <:match_case< $`int:Pa_type_conv.hash_variant cnstr$ -> `$cnstr$ >> in
             add_mc next this_mc t
         | <:ctyp< `$cnstr$ of $arg_tp$ >> ->
             atoms_only := false;
             let bnds, args_expr = handle_arg_tp _loc full_type_name arg_tp in
             let rhs = let_ins _loc bnds <:expr< `$cnstr$ $args_expr$ >> in
-            let this_mc = <:match_case< `$cnstr$ -> $rhs$ >> in
+            let this_mc = <:match_case< $`int:Pa_type_conv.hash_variant cnstr$ -> $rhs$ >> in
             add_mc next this_mc t
         | (<:ctyp< $id:_$ >> | <:ctyp< $_$ $_$ >>
         | <:ctyp< #$id:_$ >>) as inh ->
@@ -1030,7 +1040,7 @@ module Generate_bin_read = struct
           ~record:(bin_read_record full_type_name)
           ~variants:handle_variant
           ~mani:(fun _loc _tp1 -> loop _loc)
-          ~nil:(bin_read_nil type_name)
+          ~nil:(bin_read_nil full_type_name)
       in
       loop _loc rhs
     in
@@ -1060,7 +1070,8 @@ module Generate_bin_read = struct
           >>
         else
           match oc_body with
-          | `Closed expr -> <:expr< fun buf ~pos_ref -> $expr$ buf ~pos_ref >>
+          | `Closed expr ->
+            alias_or_fun expr <:expr< fun buf ~pos_ref -> $expr$ buf ~pos_ref >>
           | `Open body -> <:expr< fun buf ~pos_ref -> $body$ >>
       in
       let func = Gen.abstract _loc arg_patts body in
@@ -1084,8 +1095,9 @@ module Generate_bin_read = struct
                     rewrite_call (fun new_f -> cnv (<:expr< $new_f$ $arg$ >>)) f
                 | <:expr< Bin_prot.Read.$_$ >> -> wrong_type
                 | <:expr< $lid:name$ >> when name.[0] = '_' && name.[1] = 'o' ->
+                  (* change [buf] -> [_buf] to fix unused-var warning in generated code *)
                     <:expr<
-                      fun buf ~pos_ref _vint ->
+                      fun _buf ~pos_ref _vint ->
                         Bin_prot.Common.raise_read_error
                           (Bin_prot.Common.ReadError.Silly_type $str:full_type_name$)
                           !pos_ref
@@ -1099,9 +1111,10 @@ module Generate_bin_read = struct
                               _loc (("__" ^ call ^ "__") :: rest)$
                           >>
                         in
-                        <:expr<
+                        let cnv_expr = cnv expr in
+                        alias_or_fun cnv_expr <:expr<
                           fun buf ~pos_ref vint ->
-                            $cnv expr$ buf ~pos_ref vint
+                            $cnv_expr$ buf ~pos_ref vint
                         >>
                     | _ -> assert false)  (* impossible *)
                 | _ -> assert false  (* impossible *)
@@ -1133,14 +1146,14 @@ module Generate_bin_read = struct
         Gen.apply _loc <:expr< $lid:read_name$ >>
           tparam_read_exprs
       in
-      <:expr< fun buf ~pos_ref -> $call$ buf ~pos_ref >>
+      alias_or_fun call <:expr< fun buf ~pos_ref -> $call$ buf ~pos_ref >>
     in
     let vtag_read =
       let call =
         Gen.apply _loc <:expr< $lid:vtag_read_name$ >>
           tparam_read_exprs
       in
-      <:expr< fun buf ~pos_ref vtag -> $call$ buf ~pos_ref vtag >>
+      alias_or_fun call <:expr< fun buf ~pos_ref vtag -> $call$ buf ~pos_ref vtag >>
     in
     let reader =
       <:expr<
